@@ -1,7 +1,8 @@
 // Copyright (c) 2020-2023 Christopher Davis
-// Copyright (c) 2022-2023 Sophie Herold
-// Copyright (c) 2022 Elton A Rodrigues
+// Copyright (c) 2022-2024 Sophie Herold
 // Copyright (c) 2022 Maximiliano Sandoval R
+// Copyright (c) 2023 Julian Hofer
+// Copyright (c) 2023 Starccy
 // Copyright (c) 2023 FineFindus
 // Copyright (c) 2023 Huan Nguyen
 // Copyright (c) 2023 Philipp Kiemle
@@ -29,37 +30,35 @@
 //! [`LpSlidingView`]: crate::widgets::LpSlidingView
 //! [`LpFileModel`]: crate::file_model::LpFileModel
 
+use std::cell::{Cell, OnceCell, RefCell};
+use std::marker::PhantomData;
+
+use adw::prelude::*;
+use adw::subclass::prelude::*;
+use anyhow::Context;
+use ashpd::desktop::{wallpaper, ResponseError};
+use ashpd::{Error, WindowIdentifier};
+use glib::translate::IntoGlib;
+use glib::{clone, Properties};
+use gtk::CompositeTemplate;
+
 use crate::deps::*;
 use crate::file_model::LpFileModel;
 use crate::util::gettext::*;
 use crate::util::{Direction, Position};
 use crate::widgets::{LpImage, LpImagePage, LpPrint, LpSlidingView};
 
-use crate::util::spawn;
-use adw::prelude::*;
-use adw::subclass::prelude::*;
-use anyhow::Context;
-use ashpd::desktop::{wallpaper, ResponseError};
-use ashpd::Error;
-use ashpd::WindowIdentifier;
-use glib::clone;
-use glib::translate::IntoGlib;
-use gtk::CompositeTemplate;
-use once_cell::sync::Lazy;
-
-use std::cell::{Cell, RefCell};
-
 // The number of pages we want to buffer
 // on either side of the current page.
 const BUFFER: usize = 2;
 
 mod imp {
-    use std::cell::OnceCell;
 
     use super::*;
 
-    #[derive(Debug, Default, CompositeTemplate)]
-    #[template(file = "../../data/gtk/image_view.ui")]
+    #[derive(Debug, Default, CompositeTemplate, Properties)]
+    #[properties(wrapper_type = super::LpImageView)]
+    #[template(file = "image_view.ui")]
     pub struct LpImageView {
         /// Direct child of this Adw::Bin
         #[template_child]
@@ -74,16 +73,31 @@ mod imp {
         pub(super) controls_box_end: TemplateChild<gtk::Widget>,
         #[template_child]
         pub(super) controls_box_end_events: TemplateChild<gtk::EventControllerMotion>,
+        #[template_child]
+        pub(super) fullscreen_button: TemplateChild<gtk::Button>,
 
         #[template_child]
         pub sliding_view: TemplateChild<LpSlidingView>,
 
         pub drag_source: gtk::DragSource,
 
+        /// File where trash operation has been undone
+        #[property(get, set, nullable)]
+        trash_restore: RefCell<Option<gio::File>>,
+
         pub(super) model: RefCell<LpFileModel>,
         pub(super) preserve_content: Cell<bool>,
 
         pub(super) current_image_signals: OnceCell<glib::SignalGroup>,
+
+        #[property(get = Self::current_page)]
+        _current_page: PhantomData<Option<LpImagePage>>,
+
+        #[property(get = Self::is_previous_available)]
+        _is_previous_available: PhantomData<bool>,
+
+        #[property(get = Self::is_next_available)]
+        _is_next_available: PhantomData<bool>,
     }
 
     #[glib::object_subclass]
@@ -94,7 +108,6 @@ mod imp {
 
         fn class_init(klass: &mut Self::Class) {
             Self::bind_template(klass);
-            Self::Type::bind_template_callbacks(klass);
         }
 
         fn instance_init(obj: &glib::subclass::InitializingObject<Self>) {
@@ -104,31 +117,15 @@ mod imp {
 
     impl ObjectImpl for LpImageView {
         fn properties() -> &'static [glib::ParamSpec] {
-            static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
-                vec![
-                    glib::ParamSpecObject::builder::<LpImagePage>("current-page")
-                        .read_only()
-                        .build(),
-                    glib::ParamSpecBoolean::builder("is-previous-available")
-                        .read_only()
-                        .build(),
-                    glib::ParamSpecBoolean::builder("is-next-available")
-                        .read_only()
-                        .build(),
-                ]
-            });
-
-            PROPERTIES.as_ref()
+            Self::derived_properties()
         }
 
-        fn property(&self, _id: usize, psec: &glib::ParamSpec) -> glib::Value {
-            let obj = self.obj();
-            match psec.name() {
-                "current-page" => obj.current_page().to_value(),
-                "is-previous-available" => obj.is_previous_available().to_value(),
-                "is-next-available" => obj.is_next_available().to_value(),
-                _ => unimplemented!(),
-            }
+        fn set_property(&self, id: usize, value: &glib::Value, pspec: &glib::ParamSpec) {
+            self.derived_set_property(id, value, pspec)
+        }
+
+        fn property(&self, id: usize, pspec: &glib::ParamSpec) -> glib::Value {
+            self.derived_property(id, pspec)
         }
 
         fn constructed(&self) {
@@ -136,9 +133,17 @@ mod imp {
 
             self.parent_constructed();
 
-            // Manually mange widget layout, see `WidgetImpl` for details
+            // Manually manage widget layout, see `WidgetImpl` for details
             obj.set_layout_manager(None::<gtk::LayoutManager>);
 
+            self.sliding_view.connect_current_page_notify(
+                glib::clone!(@weak obj => move |_| obj.page_changed()),
+            );
+
+            self.sliding_view.connect_target_page_reached(
+                glib::clone!(@weak obj => move || obj.target_page_reached()),
+            );
+
             let signal_group = glib::SignalGroup::new::<LpImage>();
             obj.connect_notify_local(
                 Some("current-page"),
@@ -207,7 +212,8 @@ mod imp {
     }
 
     impl WidgetImpl for LpImageView {
-        /// This manual implementation makes sure left an right overlay both fit into the window
+        /// This manual implementation makes sure left an right overlay both fit
+        /// into the window
         fn measure(&self, orientation: gtk::Orientation, for_size: i32) -> (i32, i32, i32, i32) {
             // Measure child of AdwBin
             let (child_min, child_natural, _, _) = self.bin_child.measure(orientation, for_size);
@@ -237,6 +243,35 @@ mod imp {
     }
 
     impl BinImpl for LpImageView {}
+
+    impl LpImageView {
+        pub fn current_page(&self) -> Option<LpImagePage> {
+            self.sliding_view.current_page()
+        }
+
+        /// Returns `true` if there is an image before the current one
+        pub fn is_previous_available(&self) -> bool {
+            let obj = self.obj();
+
+            if let Some(file) = obj.current_file() {
+                obj.model().index_of(&file) != Some(0)
+            } else {
+                false
+            }
+        }
+
+        /// Returns `true` if there is an image after the current one
+        pub fn is_next_available(&self) -> bool {
+            let obj = self.obj();
+
+            if let Some(file) = obj.current_file() {
+                let model = obj.model();
+                model.index_of(&file) != Some(model.n_files().saturating_sub(1))
+            } else {
+                false
+            }
+        }
+    }
 }
 
 glib::wrapper! {
@@ -245,7 +280,6 @@ glib::wrapper! {
         @implements gtk::Buildable, gtk::ConstraintTarget, gtk::Orientable;
 }
 
-#[gtk::template_callbacks]
 impl LpImageView {
     pub fn set_images_from_files(&self, files: Vec<gio::File>) {
         // Add image to recently used file. Does not work in Flatpaks:
@@ -263,14 +297,14 @@ impl LpImageView {
     }
 
     fn load_files(&self, files: Vec<gio::File>) {
-        let sliding_view = self.sliding_view();
+        let sliding_view = self.sliding_view().editor();
 
         if let Some(first) = files.first().cloned() {
             let model = LpFileModel::from_files(files);
             self.set_model(model);
             let page = self.new_image_page(&first);
 
-            sliding_view.clear();
+            sliding_view.clear_lazy();
             sliding_view.append(&page);
             self.update_sliding_view(&first);
         } else {
@@ -306,10 +340,10 @@ impl LpImageView {
 
         // List other files in directory
         if let Some(directory) = directory {
-            spawn(glib::clone!(@weak self as obj, @strong file => async move {
+            glib::spawn_future_local(glib::clone!(@weak self as obj => async move {
                 if let Err(err) = obj.model().load_directory(directory.clone()).await {
-                    log::warn!("Failed to load directory: {err}");
-                    obj.activate_action("win.show-toast", Some(&(err.to_string(), adw::ToastPriority::High.into_glib()).to_variant()))
+                    log::warn!("Failed to load directory: {}", err.root_cause());
+                    obj.activate_action("win.show-toast", Some(&(format!("{} {}", err, err.root_cause()), adw::ToastPriority::High.into_glib()).to_variant()))
                         .unwrap();
                     return;
                 }
@@ -320,7 +354,7 @@ impl LpImageView {
     }
 
     /// Move forward or backwards
-    pub fn navigate(&self, direction: Direction) {
+    pub fn navigate(&self, direction: Direction, animated: bool) {
         if let Some(current_file) = self.current_file() {
             let new_file = match direction {
                 Direction::Forward => self.model().after(&current_file),
@@ -328,7 +362,7 @@ impl LpImageView {
             };
 
             if let Some(new_file) = new_file {
-                self.scroll_sliding_view(&new_file);
+                self.scroll_sliding_view(&new_file, animated);
             }
         }
     }
@@ -382,7 +416,7 @@ impl LpImageView {
 
         log::debug!("Scrolling to page for {}", new_file.uri());
         self.imp().preserve_content.set(true);
-        sliding_view.scroll_to(&page);
+        sliding_view.animate_to(&page);
     }
 
     /// Ensures the sliding view contains the correct images
@@ -391,7 +425,7 @@ impl LpImageView {
             "Updating sliding_view neighbors for current path {}",
             current_file.uri()
         );
-        let sliding_view = self.sliding_view();
+        let sliding_view = self.sliding_view().editor();
 
         self.imp().preserve_content.set(false);
 
@@ -401,7 +435,7 @@ impl LpImageView {
         // remove old pages
         for (uri, page) in &existing {
             if !target.contains_key(uri) {
-                sliding_view.remove(page);
+                sliding_view.remove_lazy(page);
             }
         }
 
@@ -418,8 +452,8 @@ impl LpImageView {
         self.notify("is-next-available");
     }
 
-    fn scroll_sliding_view(&self, file: &gio::File) {
-        let Some(current_page) = self.sliding_view().pages().remove(&file.uri()) else {
+    fn scroll_sliding_view(&self, file: &gio::File, animated: bool) {
+        let Some(current_page) = self.sliding_view().pages().swap_remove(&file.uri()) else {
             log::error!(
                 "Current path not available in sliding_view for scrolling: {}",
                 file.uri()
@@ -427,7 +461,11 @@ impl LpImageView {
             return;
         };
 
-        self.sliding_view().scroll_to(&current_page);
+        if animated {
+            self.sliding_view().animate_to(&current_page);
+        } else {
+            self.sliding_view().instant_to(&current_page);
+        }
     }
 
     fn sliding_view(&self) -> LpSlidingView {
@@ -492,6 +530,16 @@ impl LpImageView {
 
     /// Handle files are added or removed from directory
     fn model_content_changed_cb(&self) {
+        // Animate to image that was restored from trash
+        if let Some(trash_restore) = self.trash_restore() {
+            if self.model().contains(&trash_restore) {
+                self.set_trash_restore(gio::File::NONE);
+                self.update_sliding_view(&trash_restore);
+                self.scroll_sliding_view(&trash_restore, true);
+                return;
+            }
+        }
+
         let Some(current_file) = self.current_file() else {
             return;
         };
@@ -534,10 +582,6 @@ impl LpImageView {
             .expect("Signal group should be set up during construction")
     }
 
-    pub fn current_page(&self) -> Option<LpImagePage> {
-        self.imp().sliding_view.current_page()
-    }
-
     pub fn current_uri(&self) -> Option<glib::GString> {
         self.imp()
             .sliding_view
@@ -556,30 +600,10 @@ impl LpImageView {
         self.imp().drag_source.clone()
     }
 
-    /// Returns `true` if there is an image before the current one
-    pub fn is_previous_available(&self) -> bool {
-        if let Some(file) = self.current_file() {
-            self.model().index_of(&file) != Some(0)
-        } else {
-            false
-        }
-    }
-
-    /// Returns `true` if there is an image after the current one
-    pub fn is_next_available(&self) -> bool {
-        if let Some(file) = self.current_file() {
-            let model = self.model();
-            model.index_of(&file) != Some(model.n_files().saturating_sub(1))
-        } else {
-            false
-        }
-    }
-
-    #[template_callback]
-    fn page_changed_cb(&self) {
-        self.notify("current-page");
-        self.notify("is-previous-available");
-        self.notify("is-next-available");
+    fn page_changed(&self) {
+        self.notify_current_page();
+        self.notify_is_next_available();
+        self.notify_is_previous_available();
 
         let Some(new_page) = self.current_page() else {
             log::debug!("Page changed but no current page");
@@ -591,9 +615,8 @@ impl LpImageView {
         }
     }
 
-    #[template_callback]
     /// Called when page change animation completed
-    fn target_page_reached_cb(&self) {
+    fn target_page_reached(&self) {
         let current_page = self.current_page();
 
         if self.imp().preserve_content.get() {
@@ -607,7 +630,7 @@ impl LpImageView {
         // Reset zoom and rotation of other pages
         if let Some(new_page) = &current_page {
             let mut other_pages = self.sliding_view().pages();
-            other_pages.remove(&new_page.file().uri());
+            other_pages.swap_remove(&new_page.file().uri());
             for (_, page) in other_pages {
                 let image = page.image();
                 image.reset_rotation();
@@ -617,25 +640,14 @@ impl LpImageView {
         }
     }
 
-    #[template_callback]
-    fn get_fullscreen_icon(&self, fullscreened: bool) -> &'static str {
-        if fullscreened {
+    pub fn on_fullscreen_changed(&self, is_fullscreened: bool) {
+        let icon = if is_fullscreened {
             "view-restore-symbolic"
         } else {
             "view-fullscreen-symbolic"
-        }
-    }
+        };
 
-    pub fn zoom_out(&self) {
-        if let Some(current_page) = self.current_page() {
-            current_page.image().zoom_out();
-        }
-    }
-
-    pub fn zoom_in(&self) {
-        if let Some(current_page) = self.current_page() {
-            current_page.image().zoom_in();
-        }
+        self.imp().fullscreen_button.set_icon_name(icon);
     }
 
     pub fn rotate_image(&self, angle: f64) {
@@ -665,7 +677,8 @@ impl LpImageView {
                     "win.show-toast",
                     Some(
                         &(
-                            // Translators: This is a toast notification, informing the user that an image has been set as background.
+                            // Translators: This is a toast notification, informing the user that
+                            // an image has been set as background.
                             gettext("Set as background."),
                             adw::ToastPriority::High.into_glib(),
                         )
